@@ -368,6 +368,12 @@ lp.parseObj = function() {
   if (this.curIndent + 1 < indent) { indent = this.curIndent; line = this.curLineStart }
   while (!this.closes(tt.braceR, indent, line)) {
     let prop = this.startNode(), isGenerator, isAsync, start
+    if (this.options.ecmaVersion >= 9 && this.eat(tt.ellipsis)) {
+      prop.argument = this.parseMaybeAssign()
+      node.properties.push(this.finishNode(prop, "SpreadElement"))
+      this.eat(tt.comma)
+      continue
+    }
     if (this.options.ecmaVersion >= 6) {
       start = this.storeCurrentPos()
       prop.method = false
@@ -376,8 +382,9 @@ lp.parseObj = function() {
     }
     this.parsePropertyName(prop)
     if (this.toks.isAsyncProp(prop)) {
-      this.parsePropertyName(prop)
       isAsync = true
+      isGenerator = this.options.ecmaVersion >= 9 && this.eat(tt.star)
+      this.parsePropertyName(prop)
     } else {
       isAsync = false
     }
@@ -391,7 +398,7 @@ lp.parseObj = function() {
       prop.value = this.parseMethod(isGenerator, isAsync)
     } else if (this.options.ecmaVersion >= 5 && prop.key.type === "Identifier" &&
                !prop.computed && (prop.key.name === "get" || prop.key.name === "set") &&
-               (this.tok.type != tt.comma && this.tok.type != tt.braceR)) {
+               (this.tok.type != tt.comma && this.tok.type != tt.braceR && this.tok.type != tt.eq)) {
       prop.kind = prop.key.name
       this.parsePropertyName(prop)
       prop.value = this.parseMethod(false)
@@ -476,12 +483,13 @@ lp.toAssignable = function(node, binding) {
     return this.dummyIdent()
   } else if (node.type == "ObjectExpression") {
     node.type = "ObjectPattern"
-    let props = node.properties
-    for (let prop of props)
-      this.toAssignable(prop.value, binding)
+    for (let prop of node.properties)
+      this.toAssignable(prop, binding)
   } else if (node.type == "ArrayExpression") {
     node.type = "ArrayPattern"
     this.toAssignableList(node.elements, binding)
+  } else if (node.type == "Property") {
+    this.toAssignable(node.value, binding)
   } else if (node.type == "SpreadElement") {
     node.type = "RestElement"
     this.toAssignable(node.argument, binding)
@@ -514,8 +522,8 @@ lp.parseMethod = function(isGenerator, isAsync) {
     node.async = !!isAsync
   this.inAsync = node.async
   node.params = this.parseFunctionParams()
-  node.expression = this.options.ecmaVersion >= 6 && this.tok.type !== tt.braceL
-  node.body = node.expression ? this.parseMaybeAssign() : this.parseBlock()
+  node.body = this.parseBlock()
+  this.toks.adaptDirectivePrologue(node.body.body)
   this.inAsync = oldInAsync
   return this.finishNode(node, "FunctionExpression")
 }
@@ -528,7 +536,12 @@ lp.parseArrowExpression = function(node, params, isAsync) {
   this.inAsync = node.async
   node.params = this.toAssignableList(params, true)
   node.expression = this.tok.type !== tt.braceL
-  node.body = node.expression ? this.parseMaybeAssign() : this.parseBlock()
+  if (node.expression) {
+    node.body = this.parseMaybeAssign()
+  } else {
+    node.body = this.parseBlock()
+    this.toks.adaptDirectivePrologue(node.body.body)
+  }
   this.inAsync = oldInAsync
   return this.finishNode(node, "ArrowFunctionExpression")
 }
