@@ -1,7 +1,7 @@
 import {types as tt} from "./tokentype"
 import {Parser} from "./state"
 import {lineBreak, skipWhiteSpace} from "./whitespace"
-import {isIdentifierStart, isIdentifierChar} from "./identifier"
+import {isIdentifierStart, isIdentifierChar, keywordRelationalOperator} from "./identifier"
 import {has} from "./util"
 import {DestructuringErrors} from "./parseutil"
 
@@ -21,6 +21,7 @@ pp.parseTopLevel = function(node) {
     let stmt = this.parseStatement(true, true, exports)
     node.body.push(stmt)
   }
+  this.adaptDirectivePrologue(node.body)
   this.next()
   if (this.options.ecmaVersion >= 6) {
     node.sourceType = this.options.sourceType
@@ -31,7 +32,7 @@ pp.parseTopLevel = function(node) {
 const loopLabel = {kind: "loop"}, switchLabel = {kind: "switch"}
 
 pp.isLet = function() {
-  if (this.type !== tt.name || this.options.ecmaVersion < 6 || this.value != "let") return false
+  if (this.options.ecmaVersion < 6 || !this.isContextual("let")) return false
   skipWhiteSpace.lastIndex = this.pos
   let skip = skipWhiteSpace.exec(this.input)
   let next = this.pos + skip[0].length, nextCh = this.input.charCodeAt(next)
@@ -40,7 +41,7 @@ pp.isLet = function() {
     let pos = next + 1
     while (isIdentifierChar(this.input.charCodeAt(pos), true)) ++pos
     let ident = this.input.slice(next, pos)
-    if (!this.isKeyword(ident)) return true
+    if (!keywordRelationalOperator.test(ident)) return true
   }
   return false
 }
@@ -49,7 +50,7 @@ pp.isLet = function() {
 // - 'async /*foo*/ function' is OK.
 // - 'async /*\n*/ function' is invalid.
 pp.isAsyncFunction = function() {
-  if (this.type !== tt.name || this.options.ecmaVersion < 8 || this.value != "async")
+  if (this.options.ecmaVersion < 8 || !this.isContextual("async"))
     return false
 
   skipWhiteSpace.lastIndex = this.pos
@@ -119,7 +120,8 @@ pp.parseStatement = function(declaration, topLevel, exports) {
     // next token is a colon and the expression was a simple
     // Identifier node, we switch to interpreting it as a label.
   default:
-    if (this.isAsyncFunction() && declaration) {
+    if (this.isAsyncFunction()) {
+      if (!declaration) this.unexpected()
       this.next()
       return this.parseFunctionStatement(node, true)
     }
@@ -185,10 +187,14 @@ pp.parseDoStatement = function(node) {
 
 pp.parseForStatement = function(node) {
   this.next()
+  let awaitAt = (this.options.ecmaVersion >= 9 && this.inAsync && this.eatContextual("await")) ? this.lastTokStart : -1
   this.labels.push(loopLabel)
   this.enterLexicalScope()
   this.expect(tt.parenL)
-  if (this.type === tt.semi) return this.parseFor(node, null)
+  if (this.type === tt.semi) {
+    if (awaitAt > -1) this.unexpected(awaitAt)
+    return this.parseFor(node, null)
+  }
   let isLet = this.isLet()
   if (this.type === tt._var || this.type === tt._const || isLet) {
     let init = this.startNode(), kind = isLet ? "let" : this.value
@@ -196,20 +202,32 @@ pp.parseForStatement = function(node) {
     this.parseVar(init, true, kind)
     this.finishNode(init, "VariableDeclaration")
     if ((this.type === tt._in || (this.options.ecmaVersion >= 6 && this.isContextual("of"))) && init.declarations.length === 1 &&
-        !(kind !== "var" && init.declarations[0].init))
+        !(kind !== "var" && init.declarations[0].init)) {
+      if (this.options.ecmaVersion >= 9) {
+        if (this.type === tt._in) {
+          if (awaitAt > -1) this.unexpected(awaitAt)
+        } else node.await = awaitAt > -1
+      }
       return this.parseForIn(node, init)
+    }
+    if (awaitAt > -1) this.unexpected(awaitAt)
     return this.parseFor(node, init)
   }
   let refDestructuringErrors = new DestructuringErrors
   let init = this.parseExpression(true, refDestructuringErrors)
   if (this.type === tt._in || (this.options.ecmaVersion >= 6 && this.isContextual("of"))) {
-    this.toAssignable(init)
+    if (this.options.ecmaVersion >= 9) {
+      if (this.type === tt._in) {
+        if (awaitAt > -1) this.unexpected(awaitAt)
+      } else node.await = awaitAt > -1
+    }
+    this.toAssignable(init, false, refDestructuringErrors)
     this.checkLVal(init)
-    this.checkPatternErrors(refDestructuringErrors, true)
     return this.parseForIn(node, init)
   } else {
     this.checkExpressionErrors(refDestructuringErrors, true)
   }
+  if (awaitAt > -1) this.unexpected(awaitAt)
   return this.parseFor(node, init)
 }
 
@@ -218,16 +236,12 @@ pp.parseFunctionStatement = function(node, isAsync) {
   return this.parseFunction(node, true, false, isAsync)
 }
 
-pp.isFunction = function() {
-  return this.type === tt._function || this.isAsyncFunction()
-}
-
 pp.parseIfStatement = function(node) {
   this.next()
   node.test = this.parseParenExpression()
   // allow function declarations in branches, but only in non-strict mode
-  node.consequent = this.parseStatement(!this.strict && this.isFunction())
-  node.alternate = this.eat(tt._else) ? this.parseStatement(!this.strict && this.isFunction()) : null
+  node.consequent = this.parseStatement(!this.strict && this.type == tt._function)
+  node.alternate = this.eat(tt._else) ? this.parseStatement(!this.strict && this.type == tt._function) : null
   return this.finishNode(node, "IfStatement")
 }
 
@@ -357,11 +371,12 @@ pp.parseLabeledStatement = function(node, maybeName, expr) {
   for (let i = this.labels.length - 1; i >= 0; i--) {
     let label = this.labels[i]
     if (label.statementStart == node.start) {
+      // Update information about previous labels on this node
       label.statementStart = this.start
       label.kind = kind
     } else break
   }
-  this.labels.push({name: maybeName, kind: kind, statementStart: this.start})
+  this.labels.push({name: maybeName, kind, statementStart: this.start})
   node.body = this.parseStatement(true)
   if (node.body.type == "ClassDeclaration" ||
       node.body.type == "VariableDeclaration" && node.body.kind != "var" ||
@@ -422,8 +437,14 @@ pp.parseFor = function(node, init) {
 pp.parseForIn = function(node, init) {
   let type = this.type === tt._in ? "ForInStatement" : "ForOfStatement"
   this.next()
+  if (type == "ForInStatement") {
+    if (init.type === "AssignmentPattern" ||
+      (init.type === "VariableDeclaration" && init.declarations[0].init != null &&
+       (this.strict || init.declarations[0].id.type !== "Identifier")))
+      this.raise(init.start, "Invalid assignment in for-in loop head")
+  }
   node.left = init
-  node.right = this.parseExpression()
+  node.right = type == "ForInStatement" ? this.parseExpression() : this.parseMaybeAssign()
   this.expect(tt.parenR)
   this.exitLexicalScope()
   node.body = this.parseStatement(false)
@@ -464,7 +485,7 @@ pp.parseVarId = function(decl, kind) {
 
 pp.parseFunction = function(node, isStatement, allowExpressionBody, isAsync) {
   this.initFunction(node)
-  if (this.options.ecmaVersion >= 6 && !isAsync)
+  if (this.options.ecmaVersion >= 9 || this.options.ecmaVersion >= 6 && !isAsync)
     node.generator = this.eat(tt.star)
   if (this.options.ecmaVersion >= 8)
     node.async = !!isAsync
@@ -518,62 +539,67 @@ pp.parseClass = function(node, isStatement) {
   classBody.body = []
   this.expect(tt.braceL)
   while (!this.eat(tt.braceR)) {
-    if (this.eat(tt.semi)) continue
-    let method = this.startNode()
-    let isGenerator = this.eat(tt.star)
-    let isAsync = false
-    let isMaybeStatic = this.type === tt.name && this.value === "static"
-    this.parsePropertyName(method)
-    method.static = isMaybeStatic && this.type !== tt.parenL
-    if (method.static) {
-      if (isGenerator) this.unexpected()
-      isGenerator = this.eat(tt.star)
-      this.parsePropertyName(method)
-    }
-    if (this.options.ecmaVersion >= 8 && !isGenerator && !method.computed &&
-        method.key.type === "Identifier" && method.key.name === "async" && this.type !== tt.parenL &&
-        !this.canInsertSemicolon()) {
-      isAsync = true
-      this.parsePropertyName(method)
-    }
-    method.kind = "method"
-    let isGetSet = false
-    if (!method.computed) {
-      let {key} = method
-      if (!isGenerator && !isAsync && key.type === "Identifier" && this.type !== tt.parenL && (key.name === "get" || key.name === "set")) {
-        isGetSet = true
-        method.kind = key.name
-        key = this.parsePropertyName(method)
-      }
-      if (!method.static && (key.type === "Identifier" && key.name === "constructor" ||
-          key.type === "Literal" && key.value === "constructor")) {
-        if (hadConstructor) this.raise(key.start, "Duplicate constructor in the same class")
-        if (isGetSet) this.raise(key.start, "Constructor can't have get/set modifier")
-        if (isGenerator) this.raise(key.start, "Constructor can't be a generator")
-        if (isAsync) this.raise(key.start, "Constructor can't be an async method")
-        method.kind = "constructor"
-        hadConstructor = true
-      }
-    }
-    this.parseClassMethod(classBody, method, isGenerator, isAsync)
-    if (isGetSet) {
-      let paramCount = method.kind === "get" ? 0 : 1
-      if (method.value.params.length !== paramCount) {
-        let start = method.value.start
-        if (method.kind === "get")
-          this.raiseRecoverable(start, "getter should have no params")
-        else
-          this.raiseRecoverable(start, "setter should have exactly one param")
-      } else {
-        if (method.kind === "set" && method.value.params[0].type === "RestElement")
-          this.raiseRecoverable(method.value.params[0].start, "Setter cannot use rest params")
-      }
+    const member = this.parseClassMember(classBody)
+    if (member && member.type === "MethodDefinition" && member.kind === "constructor") {
+      if (hadConstructor) this.raise(member.start, "Duplicate constructor in the same class")
+      hadConstructor = true
     }
   }
   node.body = this.finishNode(classBody, "ClassBody")
   return this.finishNode(node, isStatement ? "ClassDeclaration" : "ClassExpression")
 }
 
+pp.parseClassMember = function(classBody) {
+  if (this.eat(tt.semi)) return null
+
+  let method = this.startNode()
+  const tryContextual = (k, noLineBreak = false) => {
+    const start = this.start, startLoc = this.startLoc
+    if (!this.eatContextual(k)) return false
+    if (this.type !== tt.parenL && (!noLineBreak || !this.canInsertSemicolon())) return true
+    if (method.key) this.unexpected()
+    method.computed = false
+    method.key = this.startNodeAt(start, startLoc)
+    method.key.name = k
+    this.finishNode(method.key, "Identifier")
+    return false
+  }
+
+  method.kind = "method"
+  method.static = tryContextual("static")
+  let isGenerator = this.eat(tt.star)
+  let isAsync = false
+  if (!isGenerator) {
+    if (this.options.ecmaVersion >= 8 && tryContextual("async", true)) {
+      isAsync = true
+      isGenerator = this.options.ecmaVersion >= 9 && this.eat(tt.star)
+    } else if (tryContextual("get")) {
+      method.kind = "get"
+    } else if (tryContextual("set")) {
+      method.kind = "set"
+    }
+  }
+  if (!method.key) this.parsePropertyName(method)
+  let {key} = method
+  if (!method.computed && !method.static && (key.type === "Identifier" && key.name === "constructor" ||
+      key.type === "Literal" && key.value === "constructor")) {
+    if (method.kind !== "method") this.raise(key.start, "Constructor can't have get/set modifier")
+    if (isGenerator) this.raise(key.start, "Constructor can't be a generator")
+    if (isAsync) this.raise(key.start, "Constructor can't be an async method")
+    method.kind = "constructor"
+  } else if (method.static && key.type === "Identifier" && key.name === "prototype") {
+    this.raise(key.start, "Classes may not have a static property named prototype")
+  }
+  this.parseClassMethod(classBody, method, isGenerator, isAsync)
+  if (method.kind === "get" && method.value.params.length !== 0)
+    this.raiseRecoverable(method.value.start, "getter should have no params")
+  if (method.kind === "set" && method.value.params.length !== 1)
+    this.raiseRecoverable(method.value.start, "setter should have exactly one param")
+  if (method.kind === "set" && method.value.params[0].type === "RestElement")
+    this.raiseRecoverable(method.value.params[0].start, "Setter cannot use rest params")
+  return method
+}
+
 pp.parseClassMethod = function(classBody, method, isGenerator, isAsync) {
   method.value = this.parseMethod(isGenerator, isAsync)
   classBody.body.push(this.finishNode(method, "MethodDefinition"))
@@ -594,7 +620,8 @@ pp.parseExport = function(node, exports) {
   // export * from '...'
   if (this.eat(tt.star)) {
     this.expectContextual("from")
-    node.source = this.type === tt.string ? this.parseExprAtom() : this.unexpected()
+    if (this.type !== tt.string) this.unexpected()
+    node.source = this.parseExprAtom()
     this.semicolon()
     return this.finishNode(node, "ExportAllDeclaration")
   }
@@ -628,7 +655,8 @@ pp.parseExport = function(node, exports) {
     node.declaration = null
     node.specifiers = this.parseExportSpecifiers(exports)
     if (this.eatContextual("from")) {
-      node.source = this.type === tt.string ? this.parseExprAtom() : this.unexpected()
+      if (this.type !== tt.string) this.unexpected()
+      node.source = this.parseExprAtom()
     } else {
       // check for keywords used as local names
       for (let spec of node.specifiers) {
@@ -655,13 +683,17 @@ pp.checkPatternExport = function(exports, pat) {
     this.checkExport(exports, pat.name, pat.start)
   else if (type == "ObjectPattern")
     for (let prop of pat.properties)
-      this.checkPatternExport(exports, prop.value)
+      this.checkPatternExport(exports, prop)
   else if (type == "ArrayPattern")
     for (let elt of pat.elements) {
       if (elt) this.checkPatternExport(exports, elt)
     }
+  else if (type == "Property")
+    this.checkPatternExport(exports, pat.value)
   else if (type == "AssignmentPattern")
     this.checkPatternExport(exports, pat.left)
+  else if (type == "RestElement")
+    this.checkPatternExport(exports, pat.argument)
   else if (type == "ParenthesizedExpression")
     this.checkPatternExport(exports, pat.expression)
 }
@@ -760,3 +792,19 @@ pp.parseImportSpecifiers = function() {
   }
   return nodes
 }
+
+// Set `ExpressionStatement#directive` property for directive prologues.
+pp.adaptDirectivePrologue = function(statements) {
+  for (let i = 0; i < statements.length && this.isDirectiveCandidate(statements[i]); ++i) {
+    statements[i].directive = statements[i].expression.raw.slice(1, -1)
+  }
+}
+pp.isDirectiveCandidate = function(statement) {
+  return (
+    statement.type === "ExpressionStatement" &&
+    statement.expression.type === "Literal" &&
+    typeof statement.expression.value === "string" &&
+    // Reject parenthesized strings.
+    (this.input[statement.start] === "\"" || this.input[statement.start] === "'")
+  )
+}
